/**
 * east工具集，要能够适用于前后端
 */

const etools={};
module.exports=etools;

// 直接引用第三方库
etools.clone=require('clone');
// 子项目载入
etools.time=require('./sub_time');
etools.num =require('./sub_num');
etools.md5 =require('./md5');

// 是否是在nodejs环境
etools.isNode=process && process.version;

// 获取对象原型类别
etools.protoType=function(obj){
  return Object.prototype.toString.call(obj);
};

// 获取east定义类别
etools.etype=function(obj){
  switch (Object.prototype.toString.call(obj)){
    case '[object Object]':return 'dict';
    case '[object Array]':return 'array';
    case '[object Null]':return 'null';
    case '[object RegExp]':return 'regex';
    case '[object Uint8Array]':return 'buffer';
    case '[object Undefined]':return 'undefined';
    case '[object Number]':return 'number';
    case '[object String]':return 'string';
    case '[object Function]':return 'function';
    case '[object Boolean]':return 'bool';
    default:return 'unknow';
  }
};

// 日志输出，增加时间前缀
etools.log=function(str){
  // 后台需补充时间
  if(etools.isNode) console.log(etools.time.nowLocalStr()+': '+str);
  // 前端直接输出
  else console.log(str);
};

// 取环境变量，取不到则用默认值
etools.getEnvInt=(key,def)=>{
  const val=process.env[key];
  if(val && val.length>0) return Number.parseInt(val);
  else return Number.parseInt(def);
};
etools.getEnvStr=(key,def)=>{
  const val=process.env[key];
  if(val && val.length>0) return val;
  else return String(def);
};

// 产生指定长度的随机字符串
etools.ranStr=(len,typ)=>{
    let $chars;
    switch (typ){
        case 'hex':case undefined:
        $chars='1234567890abcdef';break;
        case 'num':
            $chars='1234567890';break;
        case 'base62':
            $chars='1234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';break;
        default:
            throw new Error('unknow Random Type');
    }

    var maxPos = $chars.length;
    var pwd = '';
    for (var i = 0; i < len; i++) {
        pwd += $chars.charAt(Math.floor(Math.random() * maxPos));
    }
    return pwd;
};

/* 获取json的md5特征key */
function json2buf(obj){
  const bufArry=[];
  switch (etools.protoType(obj)){
    case '[object Array]':
      bufArry.push(Buffer.from('['));
      for(let i=0;i<obj.length;i++){
        if(i>0) bufArry.push(Buffer.from(','));
        bufArry.push(json2buf(obj[i]));
      }
      bufArry.push(Buffer.from(']'));
      return Buffer.concat(bufArry);
    case '[object Object]':
      const keyArry=Object.keys(obj).sort();
      bufArry.push(Buffer.from('{'));
      for(let i=0;i<keyArry.length;i++){
        if(i>0) bufArry.push(Buffer.from(','));
        const key=keyArry[i];
        bufArry.push(Buffer.from(`"${key}":`));
        bufArry.push(json2buf(obj[key]));
      }
      bufArry.push(Buffer.from('}'));
      return Buffer.concat(bufArry);
    case '[object Uint8Array]':// 即buffer
      return obj;
    case '[object Number]':
      return Buffer.from(obj.toString());
    case '[object String]':
      return Buffer.from(`"${obj}":`);
    case '[object Null]':
      return Buffer.from('null');
    case '[object Undefined]':
      return Buffer.from('undefined');
    case '[object Boolean]':
      return Buffer.from(obj.toString());
    default:
      throw new Error('unknow type: '+Object.prototype.toString.call(obj))
  }
}
etools.jsonMd5Key=(json)=>{
  // 必须是对象，否则报错
  if(etools.protoType(json)!=='[object Object]') throw new Error('json must be [object Object]');

  // 序列化为二进制流，md5加密后反馈
  return etools.md5(json2buf(json),undefined,'hex');
};

/**
 * 以指定顺序将元素插入数组，key相同则替换
 * @param arry 要插入的数组
 * @param obj 要插入的对象
 * @param key 排序属性，或数组下标
 */
etools.orderInsert=(arry,obj,key)=>{
  const len=arry.length; // 原数组长度
  const num=obj[key];    // 插入对象属性值
  let p,re;   // 最终插入位置，是否替换

  // 特殊情况快速得到结果
  if     (len===0)                [p,re]=[0,0];    // 空数组
  else if(arry[len-1][key]<num)   [p,re]=[len,0];  // 后面插入
  else if(arry[0][key]>num)       [p,re]=[0,0]  ;  // 前面插入
  else if(arry[len-1][key]===num) [p,re]=[len-1,1];// 最后一位替换
  else if(arry[0][key]===num)     [p,re]=[0,1];    // 第一位替换
  // 不满足特殊清空，二分法查找应该插入的位置
  else {
    let start=0,end=len-1,mid;
    while(start+1!==end){
      mid=((end-start)>>1)+start;
      if(arry[mid][key]===num){
        [p,re]=[mid,1];
        break;
      }
      else if(arry[mid][key]>num)
        end=mid;
      else
        start=mid;
    }
    // 若是break跳出，此处必定不成立；正常结束，必定成立
    if(start+1===end) [p,re]=[end,0];
  }

  // 对数组进行操作
  arry.splice(p,re,obj);
};

/* =========== 以下仅后端nodejs可用，前端需删除以下代码 =========== */
const fs=require('fs');
const path=require('path');

/**
 * md5加密，覆盖引入的md5.js方法
 * @param data 支持字符串或Buffer(Uint16Array)
 * @param inEncoding 字符串编码，仅data为string时有效，支持utf8(默认)、hex
 * @param outEncoding 输出格式，支持buffer(默认)、hex
 * @returns {*}
 */
etools.md5=function(data,inEncoding,outEncoding){
  const crypto=require('crypto');
  const md5 = crypto.createHash('md5');
  md5.update(data,inEncoding);
  return md5.digest(outEncoding);
};

/* 引入第三方库 */
etools.request=require('request');
etools.request_promise=(option)=>{
  return new Promise((resolve,reject)=>{
    etools.request(option,(err,...all)=>{
      if(err) reject(err);
      else resolve(...all);
    })
  })
};

/**
 * sha256加密
 * @param data 支持字符串或Buffer(Uint16Array)
 * @param inEncoding 字符串编码，仅data为string时有效，支持utf8(默认)、hex
 * @param outEncoding 输出格式，支持buffer(默认)、hex
 * @returns {*}
 */
etools.sha256=function(data,inEncoding,outEncoding){
  const crypto=require('crypto');
  const mid = crypto.createHash('sha256');
  mid.update(data,inEncoding);
  return mid.digest(outEncoding);
};

// 是否在调试模式
let isDebug=false;
process.execArgv.forEach(r=>{if(r.includes('debug')||r.includes('--inspect-brk=')) isDebug=true;});
etools.isDebug=isDebug;

// 计算机名
etools.hostname=require('os').hostname();

/* 将指定目录下的js文件映射到字典，以文件名为key */
// 载入/重载模块，非.js文件忽略
function renewAppModule(dict,dirPath,file,eventType){
  // 完整路径
  const fullpath=path.resolve(process.env.PWD,dirPath,file);
  // 验证载入过程
  let name,newModule;
  try{
    // 获取文件信息
    const fileStat=fs.statSync(fullpath);
    // 不是文件，则跳过
    if(!fileStat.isFile()) return;
    // 跳过结尾不是.js的文件
    if(file.substr(file.length-3)!=='.js') return;
    // 分离文件名
    name=file.substr(0,file.length-3);
    newModule=require(fullpath);
  }
  catch (e) {
    // 日志输出
    etools.log('[ etools.dynamicModule !Fail!  ] Module: '+name);
  }

  // 载入成功
  if(newModule){
    // 如果之前已有，则清除
    if(dict[name]) delete require.cache[require.resolve(fullpath)];
    // 赋值
    dict[name]=newModule;
    // 日志输出
    etools.log('[ etools.dynamicModule Success ] Module: '+name);
  }
}
etools.dynamicModule=(dirPath)=>{
  // 反馈的字典
  const dict={};

  // 初次遍历加载
  fs.readdirSync(dirPath).forEach(file=>renewAppModule(dict,dirPath,file));

  // 之后每隔一分钟更新一次，可靠
  setInterval(()=>fs.readdirSync(dirPath).forEach(file=>renewAppModule(dict,dirPath,file)),30000);

  // 监视文件夹变化，动态重载，在nas共享盘文件系统上无效，不可靠
  fs.watch(dirPath,{},(eventType,file)=>renewAppModule(dict,dirPath,file,eventType));

  // 返回字典对象
  return dict;
};